import sys
from pathlib import Path

import pytest


@pytest.fixture(autouse=True, scope="function")
def cleanup():
    for p in (Path(__file__).parent / "targets").iterdir():
        sys.modules.pop(f"tests.targets.{p.stem}", None)
    yield


def test_hooks():
    import tests.targets.module_prefixed as mod

    assert "bar" in mod.__all__
    assert "deprecated" in mod.__all__


@pytest.mark.parametrize(
    "monkay_fn",
    [
        lambda m: m.add_lazy_import("bar", ".fn_module:bar", no_hooks=True),
        lambda m: m.add_deprecated_lazy_import(
            "deprecated",
            {
                "path": "tests.targets.fn_module:deprecated",
                "reason": "old.",
                "new_attribute": "super_new",
            },
            no_hooks=True,
        ),
    ],
)
def test_no_hooks_collisions(monkay_fn):
    import tests.targets.module_prefixed as mod

    with pytest.raises(KeyError):
        monkay_fn(mod.monkay)


@pytest.mark.parametrize(
    "monkay_fn,export,in_all",
    [
        (lambda m: m.add_lazy_import("bar", ".fn_module:bar"), "lazy_import_bar", True),
        (
            lambda m: m.add_deprecated_lazy_import(
                "deprecated",
                {
                    "path": "tests.targets.fn_module:deprecated",
                    "reason": "old.",
                    "new_attribute": "super_new",
                },
            ),
            "deprecated_lazy_import_deprecated",
            True,
        ),
        (
            lambda m: m.add_lazy_import(
                "bar2",
                ".fn_module:bar",
                no_hooks=True,
            ),
            "bar2",
            False,
        ),
        (
            lambda m: m.add_deprecated_lazy_import(
                "deprecated2",
                {
                    "path": "tests.targets.fn_module:deprecated",
                    "reason": "old.",
                    "new_attribute": "super_new",
                },
                no_hooks=True,
            ),
            "deprecated2",
            False,
        ),
    ],
)
def test_add(monkay_fn, export, in_all):
    import tests.targets.module_prefixed as mod

    monkay_fn(mod.monkay)
    if in_all:
        assert export in mod.__all__
    assert mod.monkay.getter(export, no_warn_deprecated=True, check_globals_dict=True) is not None



from typing import Any

from monkay import Monkay


def prefix_fn(name: str, value: Any, type_: str) -> tuple[str, Any]:
    return f"{type_}_{name}", value


__all__ = {"foo", "monkay"}  # noqa
monkay = Monkay(
    globals(),
    lazy_imports={
        "bar": ".fn_module:bar",
    },
    deprecated_lazy_imports={
        "deprecated": {
            "path": "tests.targets.fn_module:deprecated",
            "reason": "old.",
            "new_attribute": "super_new",
        }
    },
    pre_add_lazy_import_hook=prefix_fn,
    post_add_lazy_import_hook=__all__.add,
)



import sys
from pathlib import Path

import pytest


@pytest.fixture(autouse=True, scope="function")
def cleanup():
    for p in (Path(__file__).parent / "targets").iterdir():
        sys.modules.pop(f"tests.targets.{p.stem}", None)
    yield


def test_settings_basic():
    import tests.targets.module_full as mod
    from tests.targets.settings import Settings, hurray

    new_settings = Settings(preloads=[], extensions=[])

    old_settings = mod.monkay.settings
    settings_path = mod.monkay._settings_definition
    assert isinstance(settings_path, str)
    assert mod.monkay.settings is old_settings
    mod.monkay.settings = new_settings
    assert mod.monkay.settings is new_settings

    mod.monkay.settings = lambda: old_settings
    assert mod.monkay.settings is old_settings
    # auto generated settings
    mod.monkay.settings = Settings
    mod.monkay.settings = "tests.targets.settings:hurray"
    assert mod.monkay.settings is hurray


def test_settings_overwrite():
    import tests.targets.module_full as mod

    old_settings = mod.monkay.settings
    settings_path = mod.monkay._settings_definition
    assert isinstance(settings_path, str)

    assert "tests.targets.module_settings_preloaded" not in sys.modules
    new_settings = old_settings.model_copy(update={"preloads": ["tests.targets.module_settings_preloaded"]})
    with mod.monkay.with_settings(new_settings) as yielded:
        assert mod.monkay.settings is new_settings
        assert mod.monkay.settings is yielded
        assert mod.monkay.settings is not old_settings
        assert "tests.targets.module_settings_preloaded" not in sys.modules
        mod.monkay.evaluate_settings()
        assert "tests.targets.module_settings_preloaded" in sys.modules

        # overwriting settings doesn't affect temporary scope
        mod.monkay.settings = mod.monkay._settings_definition
        assert mod.monkay.settings is new_settings

        # now access the non-temporary settings
        with mod.monkay.with_settings(None):
            assert mod.monkay.settings is not new_settings
            assert mod.monkay.settings is not old_settings


@pytest.mark.parametrize("mode", ["error", "replace", "keep"])
def test_settings_overwrite_evaluate_modes(mode):
    import tests.targets.module_full as mod

    with mod.monkay.with_settings(
        mod.monkay.settings.model_copy(update={"preloads": ["tests.targets.module_settings_preloaded"]})
    ) as new_settings:
        assert new_settings is not None
        if mode == "error":
            with pytest.raises(KeyError):
                mod.monkay.evaluate_settings(on_conflict=mode)
        else:
            mod.monkay.evaluate_settings(on_conflict=mode)


def test_settings_overwrite_evaluate_no_conflict():
    import tests.targets.module_full as mod

    with mod.monkay.with_settings(
        mod.monkay.settings.model_copy(
            update={
                "preloads": ["tests.targets.module_settings_preloaded"],
                "extensions": [],
            }
        )
    ) as new_settings:
        assert new_settings is not None
        mod.monkay.evaluate_settings(on_conflict="error")



from dataclasses import dataclass

from monkay import Monkay


@dataclass
class Extension:
    name: str = "default"

    def apply(self, app: Monkay) -> None:
        assert isinstance(app, Monkay)
        assert app.instance.is_fake_app
        print(f"{self.name} called")


@dataclass
class BrokenExtension1:
    name: str = "broken1"

    def apply(self, app: Monkay) -> None:
        app.ensure_extension("non-existent")


@dataclass
class BrokenExtension2:
    name: str = "broken2"

    def apply(self, app: Monkay) -> None:
        # not allowed here
        app.apply_extensions()


@dataclass
class NonExtension:
    name: str



# SPDX-FileCopyrightText: 2024-present alex <devkral@web.de>
#
# SPDX-License-Identifier: BSD-3-Clauses

from .base import (
    PRE_ADD_LAZY_IMPORT_HOOK,
    DeprecatedImport,
    ExtensionProtocol,
    Monkay,
    load,
    load_any,
)

__all__ = [
    "Monkay",
    "SortedExportsEntry",
    "DeprecatedImport",
    "PRE_ADD_LAZY_IMPORT_HOOK",
    "ExtensionProtocol",
    "load",
    "load_any",
    "absolutify_import",
]



import contextlib
import sys
from io import StringIO
from pathlib import Path

import pytest
from pydantic_settings import BaseSettings

from monkay import Monkay, load, load_any


@pytest.fixture(autouse=True, scope="function")
def cleanup():
    for p in (Path(__file__).parent / "targets").iterdir():
        sys.modules.pop(f"tests.targets.{p.stem}", None)
    yield


def test_preloaded():
    assert "tests.targets.module_full" not in sys.modules
    import tests.targets.module_full as mod

    assert "tests.targets.fn_module" not in sys.modules

    assert "tests.targets.module_full" in sys.modules
    assert "tests.targets.module_full_preloaded1" in sys.modules
    assert "tests.targets.module_full_preloaded1_fn" in sys.modules
    assert "tests.targets.module_preloaded1" in sys.modules
    assert "tests.targets.extension" in sys.modules

    with contextlib.redirect_stdout(StringIO()):
        mod.bar  # noqa

    assert "tests.targets.fn_module" in sys.modules


def test_attrs():
    import tests.targets.module_full as mod

    assert isinstance(mod.monkay, Monkay)

    # in extras
    assert mod.foo() == "foo"
    assert mod.bar() == "bar"
    assert mod.bar2() == "bar2"
    with pytest.raises(KeyError):
        mod.monkay.add_lazy_import("bar", "tests.targets.fn_module:bar")
    with pytest.raises(KeyError):
        mod.monkay.add_deprecated_lazy_import(
            "bar",
            {
                "path": "tests.targets.fn_module:deprecated",
                "reason": "old.",
                "new_attribute": "super_new",
            },
        )

    assert isinstance(mod.settings, BaseSettings)
    with pytest.warns(DeprecationWarning) as record:
        assert mod.deprecated() == "deprecated"
    assert record[0].message.args[0] == 'Attribute: "deprecated" is deprecated.\nReason: old.\nUse "super_new" instead.'


def test_load():
    assert load("tests.targets.fn_module.bar") is not None
    assert load("tests.targets.fn_module:bar") is not None
    with pytest.raises(ValueError):
        assert load("tests.targets.fn_module.bar", allow_splits=":") is not None
    with pytest.raises(AttributeError):
        assert load("tests.targets.fn_module:bar", allow_splits=".") is not None


def test_load_any():
    assert load_any("tests.targets.fn_module", ["not_existing", "bar"]) is not None
    with pytest.warns(DeprecationWarning) as records:
        assert (
            load_any(
                "tests.targets.fn_module",
                ["not_existing", "bar"],
                non_first_deprecated=True,
            )
            is not None
        )
    assert (
        load_any(
            "tests.targets.fn_module",
            ["bar", "not_existing"],
            non_first_deprecated=True,
        )
        is not None
    )
    assert str(records[0].message) == '"bar" is deprecated, use "not_existing" instead.'
    with pytest.raises(ImportError):
        assert load_any("tests.targets.fn_module", ["not-existing"]) is None
    with pytest.raises(ImportError):
        assert load_any("tests.targets.fn_module", []) is None
    with pytest.raises(ImportError):
        load_any("tests.targets.not_existing", ["bar"])


def test_extensions(capsys):
    import tests.targets.module_full as mod
    from tests.targets.extension import Extension, NonExtension

    captured = capsys.readouterr()
    assert captured.out == captured.err == ""

    app = mod.FakeApp()
    mod.monkay.set_instance(app)
    captured_out = capsys.readouterr().out
    assert captured_out == "settings_extension1 called\nsettings_extension2 called\n"
    with pytest.raises(ValueError):
        mod.monkay.add_extension(NonExtension(name="foo"))  # type: ignore
    with pytest.raises(KeyError):
        mod.monkay.add_extension(Extension(name="settings_extension1"))
    assert capsys.readouterr().out == ""

    # order

    class ExtensionA:
        name: str = "A"

        def apply(self, monkay: Monkay) -> None:
            monkay.ensure_extension("B")
            with pytest.raises(RuntimeError):
                monkay.ensure_extension("D")
            print("A")

    class ExtensionB:
        name: str = "B"

        def apply(self, monkay: Monkay) -> None:
            monkay.ensure_extension("A")
            monkay.ensure_extension(ExtensionC())
            print("B")

    class ExtensionC:
        name: str = "C"

        def apply(self, monkay: Monkay) -> None:
            monkay.ensure_extension(ExtensionA())
            print("C")

    with mod.monkay.with_extensions({"B": ExtensionB(), "A": ExtensionA()}):
        mod.monkay.apply_extensions()

    assert capsys.readouterr().out == "A\nC\nB\n"
    with mod.monkay.with_extensions(
        {
            "C": ExtensionC(),
            "B": ExtensionB(),
        }
    ):
        mod.monkay.apply_extensions()

    assert capsys.readouterr().out == "B\nA\nC\n"


def test_app(capsys):
    import tests.targets.module_full as mod

    app = mod.FakeApp()
    mod.monkay.set_instance(app)
    assert mod.monkay.instance is app
    captured_out = capsys.readouterr().out
    assert captured_out == "settings_extension1 called\nsettings_extension2 called\n"
    app2 = mod.FakeApp()
    with mod.monkay.with_instance(app2):
        assert mod.monkay.instance is app2
        assert capsys.readouterr().out == ""
    assert capsys.readouterr().out == ""


def test_caches():
    import tests.targets.module_full as mod

    assert not mod.monkay._cached_imports

    assert mod.bar() == "bar"
    assert "bar" in mod.monkay._cached_imports
    assert isinstance(mod.settings, BaseSettings)
    assert "settings" not in mod.monkay._cached_imports
    # settings cache
    assert "_loaded_settings" in mod.monkay.__dict__
    mod.monkay.clear_caches()

    assert not mod.monkay._cached_imports
    assert "_loaded_settings" not in mod.monkay.__dict__


def test_sorted_exports():
    import tests.targets.module_full as mod

    eval(mod.stringify_all_plain(separate_by_category=True))
    eval(mod.stringify_all_plain(separate_by_category=False))
    assert (
        mod.stringify_all(separate_by_category=True)
        == '__all__ = [\n"deprecated"\n,"bar2"\n,"bar"\n,"dynamic"\n,"settings"\n,"foo"\n,"stringify_all"\n]'
    )
    assert (
        mod.stringify_all(separate_by_category=False)
        == '__all__ = [\n"bar2"\n,"bar"\n,"deprecated"\n,"dynamic"\n,"foo"\n,"settings"\n,"stringify_all"\n]'
    )



import sys
from pathlib import Path

import pytest


@pytest.fixture(autouse=True, scope="function")
def cleanup():
    for p in (Path(__file__).parent / "targets").iterdir():
        sys.modules.pop(f"tests.targets.{p.stem}", None)
    yield


def test_find_missing():
    import tests.targets.module_full as mod

    # __all__ is autogenerated
    assert not mod.monkay.find_missing(all_var=mod.__all__, search_pathes=["tests.targets.fn_module"])
    # we can also use bools
    assert not mod.monkay.find_missing(all_var=True, search_pathes=["tests.targets.fn_module"])
    assert mod.monkay.find_missing(
        search_pathes=["tests.targets.not_existing", "tests.targets.module_preloaded1"],
    ) == {
        "tests.targets.not_existing": {"search_path_import"},
        "tests.targets.module_preloaded1.not_included_export": {"search_path_extra"},
    }
    assert mod.monkay.find_missing(all_var={}, search_pathes=["tests.targets.module_full_preloaded1"]) == {
        "bar": {"not_in_all_var"},
        "bar2": {
            "not_in_all_var",
        },
        "dynamic": {"not_in_all_var"},
        "settings": {"not_in_all_var"},
        "deprecated": {"not_in_all_var"},
        "tests.targets.module_full_preloaded1": {
            "missing_all_var",
        },
    }
    assert mod.monkay.find_missing(search_pathes=["tests.targets.module_missing_all"]) == {
        "tests.targets.module_missing_all": {"missing_all_var"},
    }

    assert mod.monkay.find_missing(search_pathes=["tests.targets.module_broken_all"]) == {
        "tests.targets.module_broken_all.broken": {"missing_attr", "search_path_extra"},
    }


def test_find_missing_no_all():
    from .targets import module_missing_all as mod

    assert mod.monkay.find_missing() == {
        "bar": {"not_in_all_var"},
        "tests.targets.module_missing_all": {"missing_all_var"},
    }


def test_find_missing_shadowed():
    from .targets import module_shadowed as mod

    assert mod.monkay.find_missing() == {
        "bar": {"shadowed"},
    }
